module vga_ball(
    input  logic        clk,
    input  logic        reset,
    input  logic [31:0] writedata,
    input  logic        write,
    input  logic        read,
    input  logic        chipselect,
    input  logic [8:0]  address,
    output logic [31:0] readdata,
    output logic [7:0]  VGA_R, VGA_G, VGA_B,
    output logic        VGA_CLK, VGA_HS, VGA_VS,
    output logic        VGA_BLANK_n, VGA_SYNC_n
);


  // ------------------------------------------------------------------
-----
  parameter DEPTH     = 256;                      // max line points
  parameter BASE_ADDR = 9'd2;                    // first valid CPU
address
  parameter LAST_ADDR = BASE_ADDR + (DEPTH*2) - 1;

  // Line buffer arrays=
  logic [31:0] line_lo [0:DEPTH-1];
  logic [31:0] line_hi [0:DEPTH-1];

  // clog basically finds the cieling of powers of two bits for a
certain number
  logic [$clog2(DEPTH*2)-1:0] wr_addr;
  logic                       we;
  logic [$clog2(DEPTH)-1:0]   wr_index;
  logic                       wr_lo;


  // A small helper to keep the main always_comb tidy
  logic address_in_range;

  // Compute once, reuse everywhere
  assign address_in_range = (address >= BASE_ADDR)
                        && (address <= LAST_ADDR);

  always_comb begin
    // defaults
    we       = 1'b0;
    wr_addr  = '0;
    wr_index = '0;
    wr_lo    = 1'b0;

    // only drive these signals when we’re actually writing
    if (chipselect && write && address_in_range) begin
      we      = 1'b1;
      // 0-based offset into our 2-words-per-line region and since =
each line takes two words we divide by two for the index
      wr_addr = address - BASE_ADDR;
      wr_index = wr_addr >> 1;

      // LSB tells us low-word (even) vs high-word (odd)
      // true goes low  (even addresses)
      // false goes high (odd addresses)
      wr_lo = (wr_addr[0] == 1'b0);
    end
  end
    // if its a low element, then we write to the line_lo buffer
  always_ff @(posedge clk) begin
    if (we) begin
      if (wr_lo)
        line_lo[wr_index] <= writedata;
      else
        line_hi[wr_index] <= writedata;
    end
  end


  //flipping the last bit should just swap from even to odd
  always_comb begin
    if (chipselect && read && (address >= BASE_ADDR) && (address <=
LAST_ADDR)) begin
      if (~wr_addr[0])
        readdata = line_lo[wr_addr[$clog2(DEPTH*2)-1:1]];
      else
        readdata = line_hi[wr_addr[$clog2(DEPTH*2)-1:1]];
    end else
      readdata = 32'd0;
  end

  // ------------------------------------------------------------------
-----

   logic [10:0] hcount;
   logic [9:0]  vcount;

   // instantiate VGA timing generator
   vga_counters counters(
     .clk50(clk), .reset(reset),
     .hcount(hcount), .vcount(vcount),
     .VGA_CLK(VGA_CLK), .VGA_HS(VGA_HS),
     .VGA_VS(VGA_VS), .VGA_BLANK_n(VGA_BLANK_n),
     .VGA_SYNC_n(VGA_SYNC_n)
   );


   logic [31:0] dx;
   logic [31:0] dy;

   always_comb begin
     // default background = black
     VGA_R = 8'd0;
     VGA_G = 8'd0;
     VGA_B = 8'd0;

     dx = (hcount >> 1) - 320;
     dy = vcount;

     // concentric semi-circles in white
     if (dx > 0) begin
       // inner band
       if ((dx*dx + dy*dy) <= 10600 && (dx*dx + dy*dy) >= 10000)
         {VGA_R, VGA_G, VGA_B} = 24'hffffff;
       // middle band
       if ((dx*dx + dy*dy) <= 20600 && (dx*dx + dy*dy) >= 20000)
         {VGA_R, VGA_G, VGA_B} = 24'hffffff;
       // middle band
       if ((dx*dx + dy*dy) <= 30600 && (dx*dx + dy*dy) >= 30000)
         {VGA_R, VGA_G, VGA_B} = 24'hffffff;
       // outer band
       if ((dx*dx + dy*dy) <= 40600 && (dx*dx + dy*dy) >= 40000)
         {VGA_R, VGA_G, VGA_B} = 24'hffffff;
     end

     // bounds for this row: rotating line in white, edges in gray

     if (vcount < DEPTH) begin
       if (((hcount >> 1) >= line_lo[vcount]) && ((hcount >> 1) <=
line_hi[vcount])) begin
         if ((vcount > 100 && vcount < 110) && (315 < hcount && 325 >
hcount)) begin
           VGA_R = 8'd255;
           VGA_G = 8'd0;
           VGA_B = 8'd0;
         end else begin
           VGA_R = 8'd255;
           VGA_G = 8'd255;
           VGA_B = 8'd255;
        end
       end
       // 1-pixel gray highlights at the edges
        else if (
             ( (hcount>>1) == line_lo[vcount] - 1 ||
               (hcount>>1) == line_lo[vcount] - 2 ||
               (hcount>>1) == line_hi[vcount] + 1 ||
               (hcount>>1) == line_hi[vcount] + 2
             )
           ) begin
         VGA_R = 8'd128;
         VGA_G = 8'd128;
         VGA_B = 8'd128;
       end
     end
   end


endmodule

module vga_counters(
 input logic         clk50, reset,
 output logic [10:0] hcount,  // hcount[10:0] is pixel column
 output logic [9:0]  vcount,  // vcount[9:0] is pixel row
 output logic        VGA_CLK, VGA_HS, VGA_VS, VGA_BLANK_n, VGA_SYNC_n);

/*
 * 640 X 480 VGA timing for a 50 MHz clock: one pixel every other cycle
 *
 * HCOUNT 1599 0             1279       1599 0
 *             _______________              ________
 * ___________|    Video      |____________|  Video
 *
 *
 * |SYNC| BP |<-- HACTIVE -->|FP|SYNC| BP |<-- HACTIVE
 *       _______________________      _____________
 * |____|       VGA_HS          |____|
 */
   // Parameters for hcount
   parameter HACTIVE      = 11'd 1280,
             HFRONT_PORCH = 11'd 32,
             HSYNC        = 11'd 192,
             HBACK_PORCH  = 11'd 96,   
             HTOTAL       = HACTIVE + HFRONT_PORCH + HSYNC +
                            HBACK_PORCH; // 1600

   // Parameters for vcount
   parameter VACTIVE      = 10'd 480,
             VFRONT_PORCH = 10'd 10,
             VSYNC        = 10'd 2,
             VBACK_PORCH  = 10'd 33,
             VTOTAL       = VACTIVE + VFRONT_PORCH + VSYNC +
                            VBACK_PORCH; // 525

   logic endOfLine;

   always_ff @(posedge clk50 or posedge reset)
     if (reset)          hcount <= 0;
     else if (endOfLine) hcount <= 0;
     else                hcount <= hcount + 11'd 1;

   assign endOfLine = hcount == HTOTAL - 1;

   logic endOfField;

   always_ff @(posedge clk50 or posedge reset)
     if (reset)          vcount <= 0;
     else if (endOfLine)
       if (endOfField)   vcount <= 0;
       else              vcount <= vcount + 10'd 1;

   assign endOfField = vcount == VTOTAL - 1;

   // Horizontal sync: from 0x520 to 0x5DF (0x57F)
   // 101 0010 0000 to 101 1101 1111
   assign VGA_HS = !( (hcount[10:8] == 3'b101) &
                      !(hcount[7:5] == 3'b111));
   assign VGA_VS = !( vcount[9:1] == (VACTIVE + VFRONT_PORCH) / 2);

   assign VGA_SYNC_n = 1'b0; // For putting sync on the green signal;
unused

   // Horizontal active: 0 to 1279     Vertical active: 0 to 479
   // 101 0000 0000  1280              01 1110 0000  480
   // 110 0011 1111  1599              10 0000 1100  524
   assign VGA_BLANK_n = !( hcount[10] & (hcount[9] | hcount[8]) ) &
                        !( vcount[9] | (vcount[8:5] == 4'b1111) );

   /* VGA_CLK is 25 MHz
    *             __    __    __
    * clk50    __|  |__|  |__|
    *       
    *             _____       __
    * hcount[0]__|     |_____|
    */
   assign VGA_CLK = hcount[0]; // 25 MHz clock: rising edge sensitive

endmodule
